#!/usr/bin/env python
"""
Automated framework for building a chaos expansion using DAKOTA.

"""
from __future__ import print_function

import logging
logging.basicConfig(level=logging.INFO, format="%(message)s")

import os
import pickle
import shutil
import sys
import time

from importlib import import_module
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from subprocess import call, check_output

parser = ArgumentParser(description=__doc__,
                        formatter_class=RawDescriptionHelpFormatter)
parser.add_argument("exp_name", type='str',
                    help="Name of derived chaos expansion")
parser.add_argument("--parallel", action="store_true",
                    help="Use asynchronous forking interface for "
                         "parallel invocations of model driver.")
parser.add_argument("--mpi", action="store_true",
                    help="Run DAKOTA with MPICH for analysis.")
parser.add_argument("-g", "--graphics", action="store_true",
                    help="Enable DAKOTA graphics during runtime")

if name == "__main__":

    args = parser.parse_args()

    # Load in the configuration based on the expansion name
    try:
        cfg_module = "{}_config".format(args.exp_name)
        cfg = import_module(cfg_module)
    except ImportError:
        logging.FATAL("Couldn't load configuration for experiment " +
                      args.exp_name)
        sys.exit()

    variables = cfg.variables
    # Number the variables
    for i, v in enumerate(variables):
        v.insert(2, i + 1)

    pce_directive = cfg.pce_directive
    param_name = cfg.pce_param_name
    param_vals = cfg.pce_param_vals

    responses = cfg.responses
    imports = cfg.imports
    function_name = cfg.function_name
    function = cfg.function

    ####################################################################
    # Save the experiment setup.
    pickle.dump(dict(exp_name=args.exp_name,
                     param_name=param_name, param_vals=param_vals,
                     variables=variables,
                     responses=responses,
                     function_name=function_name,
                     directive_base=pce_directive),
                open("%s_exp.dict" % args.exp_name, 'wb'))

    ####################################################################
    # Generate the model python and DAKOTA driver scripts and run DAKOTA
    # These scripts will initially reside in the main directory, but will
    # be copied to the `save` archive timestamped folder for referal later.

    logging.info("iterating parameter {name} over vals {vals}"
                    .format(name=param_name, vals=param_vals))

    path = sys.path
    cwd = os.getcwd()
    if not (cwd in path):
        sys.path.append(cwd)

    results_dict = {}
    for i, param_val in enumerate(param_vals):
        logging.info("   " + str(param_val))

        directive = pce_directive.format(name=param_name,
                                         val=param_val)

        pickle.dump(dict(variables=variables, responses=responses,
                         pce_directive=directive, imports=imports,
                         function_name=function_name,
                         function=function, graphics=args.graphics,
                         parallel=args.parallel),
                    open('config.p', 'wb'))

        # Generate scripts
        call(['python', 'gen_scripts.py'])
        # Set permissions
        call(['chmod', '+x', 'model_run.py'])

        # Quickly compute a save directory name using BASH `date` function
        foldername = check_output(['date',  '+%Y%m%d_%H%M%S']).strip()
        logging.info("Save folder - " + foldername)
        results_dict["%s_%r" % (param_name, param_val)] = foldername

        logging.info("Now executing simulation with DAKOTA...")
        if i > 0:
            # Take advantage of existing restart file and cached
            # fn evals
            logging.info("   reading old restart file")
            if argss.mpi:
                call(["mpiexec", "-np", "10",  "dakota",
                      "-read_restart", "dakota.rst.all",
                      "-i", "model_pce.in",
                      ">", "model_pce.out")
            else:
                get_ipython().system(u'dakota -read_restart dakota.rst.all -i model_pce.in > model_pce.out')
            print "concatenating results into restart file"
            get_ipython().system(u'dakota_restart_util cat dakota.rst dakota.rst.all dakota.rst.new')
            get_ipython().system(u'mv dakota.rst.new dakota.rst.all')
        else:
            if USE_MPI:
                get_ipython().system(u'mpiexec -np 10 dakota -i model_pce.in > model_pce.out')
            else:
                get_ipython().system(u'dakota -i model_pce.in > model_pce.out')
            get_ipython().system(u'cp dakota.rst dakota.rst.all')

        print "complete.\n"

        print "Copying files..."
        save_dir = os.path.join("save", foldername)
        os.mkdir(save_dir)

        for fn in ["model_run.py",
                   "model_pce.in", "model_pce.out", "model_pce.dat",
                   "dakota.rst", "config.p"]:
            shutil.copy2(fn, save_dir)
            print fn, "->", save_dir

        print "... done."

        time.sleep(1) # to ensure we don't over-write folders!

    pickle.dump(results_dict, open("%s_results.dict" % exp_name, 'wb'))


    # ### Process output
    #
    # Analyze the DAKOTA output log to save the chaos expansions, sobol indices,
    # etc.

    # In[ ]:

    ## Process the output of all the runs
    results_dict = pickle.load(open("%s_results.dict" % exp_name, 'rb'))

    for run_name, foldername in results_dict.iteritems():
        print run_name, foldername
        ## Generate output for sobol, pce
        get_ipython().magic(u'run process_output.py {"save/%s" % foldername}')


    # # ### Sampling
    # #
    # # Conduct a global analysis using LHS and save the results. The cell has been re-written to automate sampling over all the chaos expansions, so only run it after all of the chaos expansions have been generated. In general, you should follow these steps:
    # #
    # # 1. Re-set the IPython cluster.
    # # 2. Run the cell once with `n = 10`
    # # 3. Set `n` to the actual amount and run it.
    # #
    # # If you make changes to any of the utility toolkits (`dist_tools.py`, etc), you may need to reset the entire IPython kernel.

    # # In[5]:

    # n = 10000
    # ref_exp = "SM_OLS"
    # project_arg = "--project"
    # get_ipython().magic(u'run pce_sample --params --parcel {project_arg} -n {n} --parallel --recompute {ref_exp}')
    # ref_design = "%s_LHS_design.csv" % ref_exp

    # ## Reference to previos design available
    # for exp in ['SM_LARS', 'SM_LASSO']:
    # #for exp in ['SM_LASSO']:
    #     new_design = "%s_LHS_design.csv" % exp
    #     get_ipython().system(u'cp {ref_design} {new_design}')
    #     get_ipython().magic(u'run pce_sample --reference {new_design} -n {n} --parallel --recompute {exp}')

