"""
Hand-crafted wrapper of PCM results derived using DAKOTA.

Version: 4/27/2015

    1) The `__main__` portion of this script will process the PCE derivation output
       and create a binary file with the coefficient tables. It'll have to be ported
       with the parameterization script.

    2) Only the listed runs and orders will be saved and made available.

    TODO:
    - Modify to separately extract Nkn, Neq, and Smax

"""

import argparse
import h5py
import numpy as np
import os
import pickle

from parcel_model.activation import _unpack_aerosols, lognormal_activation

from scipy.special import orthogonal as op

from numba import jit

DEBUG = False
BLOCK = False # Block from the 'main' program being run
TEST  = False

parser = argparse.ArgumentParser(
    description="Module for quickly evaluating derived chaos expansions"
)
parser.add_argument("--DEBUG", action='store_true',
                    help="Enable debug output during setup")
parser.add_argument("--BLOCK", action='store_true',
                    help="Prevent chaos expansions from being re-derived")
parser.add_argument("--TEST", action='store_true',
                    help="Run quick test calculation")

STORE_FN = "pcm_param.h5"

RUNS = {
    "MARC_main": [3, 4, ],
    "MARC_gCCN": [3, 4, ],
}

RUNS_SAVE = {}

try:
    store = h5py.File(STORE_FN, "r")    ## Load data into memory
    for pcm, levels in RUNS.iteritems():
        save_dict = {}
        for level in levels:
            level = "expansion_order_"+str(level)
            ## Access the necessary PCM parameters from the stored data
            group = store["%s/%s" % (pcm, level)]
            coeffs = np.array(group['coeffs'][:])
            orders = np.array(group['orders'][:])
            max_orders = np.array(group['max_orders'][:])
            save_dict[level] = {'coeffs':coeffs, 'max_orders':max_orders, 'orders':orders}
        RUNS_SAVE[pcm] = save_dict
    store.close()
except IOError:
    print "WARNING: Could not open %s" % STORE_FN

def uni_to_uni(x, ai, bi, af=-1., bf=1.):
    """ Transform a uniform random variable to one with another set
    of lower/upper bounds.

    Parameters
    ----------
    x : float
        coordinate in original uniform variable space
    ai, bi : float
        lower and upper bounds of the original uniform distribution
    af, bf : float
        lower and upper bounds of the destination uniform distribution

    Returns
    -------
    float, random variable in new uniform distribution
    """
    if x < ai: return af
    if x > bi: return bf
    return ((bf-af)/(bi-ai))*(x - ai) + af


def project_gCCN(x):
    """
    Takes a vector with the parameters used in the PCE and projects them to
    the vector space spanned by the orthogonal polynomial basis.

    Parameters
    ----------
    x : [logN_ACC, logN_MOS, logN_MBS, logN_DST01, logN_DST02, logN_SSLT01,
         logmu_ACC, logmu_MOS, logmu_MBS, kappa_MOS, logV, T, P]

    Returns
    -------
    z : x projected into the vector space spanned by Legendre polynomials

    .. note:: see file docstring for more information on the vector space
    """

    bnds = [
        [-3., 4.], # logN_ACC
        [-5., 4.], # logN_MOS
        [-5., 4.], # logN_MBS
        [-5., 2.], # logN_DST01
        [-5., 1.], # logN_DST02
        [-5., 1.], # logN_SSLT01
        [-3., 0.], # logmu_ACC
        [-3., -1.], # logmu_MOS
        [-3., -1.], # logmu_MBS
        [0.1, 0.6], # kappa_MOS
        [-2., 1.0], # logV
        [240., 310.], # T
        [50000., 105000.], # P
    ]

    y = []
    for i, (xi, (ai, bi)) in enumerate(zip(x, bnds)):
        y.append(uni_to_uni(xi, ai, bi))
    return np.array(y)


def project_main(x):
    """
    Takes a vector with the parameters used in the PCE and projects them to
    the vector space spanned by the orthogonal polynomial basis.

    Parameters
    ----------
    x : [logN_ACC, logN_MOS, logN_MBS, logmu_ACC, logmu_MOS,
         logmu_MBS, kappa_MOS, logV, T, P]

    Returns
    -------
    z : x projected into the vector space spanned by Legendre polynomials

    .. note:: see file docstring for more information on the vector space
    """

    bnds = [
        [-3., 4.], # logN_ACC
        [-5., 4.], # logN_MOS
        [-5., 4.], # logN_MBS
        [-3., 0.], # logmu_ACC
        [-3., -1.], # logmu_MOS
        [-3., -1.], # logmu_MBS
        [0.1, 0.6], # kappa_MOS
        [-2., 1.0], # logV
        [240., 310.], # T
        [50000., 105000.], # P
    ]

    y = []
    for i, (xi, (ai, bi)) in enumerate(zip(x, bnds)):
        y.append(uni_to_uni(xi, ai, bi))
    return np.array(y)

@jit("f8(i8[:,:],f8[:],f8[:,:])")
def poly_eval(orders, coeffs, zin):
    nterms = orders.shape[0]
    nvars = orders.shape[1]

    evaluation = 0.
    for row in xrange(nterms):
        row_eval = 1.
        for col in xrange(nvars):
            #print (row+1, col+1, orders[row, col], zin[col, orders[row, col]])
            row_eval *= zin[col, orders[row, col]]
        #print (row+1, row_eval)
        evaluation += row_eval*coeffs[row]

    return evaluation

def param(V, T, P, aerosols=[], accom=0.1,
          mus=[], sigmas=[], Ns=[], kappas=[],
          pcm=RUNS.keys()[0], level="expansion_order_4",):

    # Is this the giant-CCN version?
    gCCN = "gCCN" in pcm

    if aerosols:
        d = _unpack_aerosols(aerosols)
    else:
        ## Assert that the aerosol was already decomposed into component vars
        assert mus
        assert sigmas
        assert Ns
        assert kappas

    ACC, MOS, MBS = aerosols[:3]
    x = [np.log10(ACC.distribution.N),
         np.log10(MOS.distribution.N),
         np.log10(MBS.distribution.N), ]
    if gCCN:
        DST01, DST02, SSLT01 = aerosols[3:]
        x.extend([np.log10(DST01.distribution.N),
                  np.log10(DST02.distribution.N),
                  np.log10(SSLT01.distribution.N),])
    x.extend([np.log10(ACC.distribution.mu),
              np.log10(MOS.distribution.mu),
              np.log10(MBS.distribution.mu),
              MOS.kappa, np.log10(V), T, P  ])

    ## Project into orthogonal polynomial space
    if gCCN:
        z = project_gCCN(x[:])
    else:
        z = project_main(x[:])

    coeffs = RUNS_SAVE[pcm][level]['coeffs']
    orders = RUNS_SAVE[pcm][level]['orders']
    max_orders = RUNS_SAVE[pcm][level]['max_orders']

    #nterms = len(coeffs)
    #nvars = len(z)

    ## Step 1) Evaluate the necessary polynomial orders
    poly_evals = []
    for zi, order in zip(z, max_orders):
        poly_evals.append([op.eval_legendre(n, zi) for n in xrange(order+1)])
    poly_evals = np.array(poly_evals)

    evaluation = poly_eval(orders, coeffs, poly_evals)

    ## Step 3) This gives us log10(Smax), so now we return Smax
    Smax = 10.**(evaluation)

    n_acts, act_fracs = [], []
    for mode in [ACC, MOS, MBS]:
        mu = mode.distribution.mu
        sigma = mode.distribution.sigma
        N = mode.distribution.N
        kappa = mode.kappa
        n_act, act_frac = lognormal_activation(Smax, mu*1e-6, sigma, N, kappa, T=T)
        n_acts.append(n_act)
        act_fracs.append(act_frac)

    return Smax, n_act, act_frac

if __name__ == "__main__":

    args = parser.parse_args()

    if args.TEST:
        import parcel_model as pm
        T = 283.
        P = 85000.


        ACC = pm.AerosolSpecies('ACC', pm.Lognorm(mu=0.17, sigma=2.0, N=50.),
                                kappa=0.507, bins=200)
        MOS = pm.AerosolSpecies('MOS', pm.Lognorm(mu=0.013, sigma=1.59, N=2100.),
                                kappa=0.28, bins=200)
        MBS = pm.AerosolSpecies('MBS', pm.Lognorm(mu=0.010, sigma=1.59, N=3920.),
                                kappa=0.507, bins=200)

        DST01  = pm.AerosolSpecies('MBS', pm.Lognorm(mu=0.16, sigma=1.99, N=500.),
                                   kappa=0.14, bins=50)
        DST02  = pm.AerosolSpecies('MBS', pm.Lognorm(mu=0.010, sigma=1.59, N=50.),
                                   kappa=0.14, bins=50)
        SSLT01 = pm.AerosolSpecies('MBS', pm.Lognorm(mu=0.010, sigma=1.59, N=80.),
                                   kappa=1.16, bins=50)

        aerosols = [ACC, MOS, MBS]

        print "MAIN"
        print "----"
        for V in [0.2, 0.5, 1.0, 2.0]:

            print "V", V
            print "PCM", param(V, T, P, aerosols, accom=0.1)
            print "ARG", pm.arg2000(V, T, P, aerosols, accom=0.1)
            print "MBN", pm.mbn2014(V, T, P, aerosols, accom=0.1)
            print "--"*30

        print
        print "gCCN"
        print "----"

        aerosols.extend([DST01, DST02, SSLT01])

        for V in [0.2, 0.5, 1.0, 2.0]:

            print "V", V
            print "PCM", param(V, T, P, aerosols, accom=0.1,
                               pcm="MARC_gCCN")
            print "ARG", pm.arg2000(V, T, P, aerosols, accom=0.1)
            print "MBN", pm.mbn2014(V, T, P, aerosols, accom=0.1)
            print "--"*30

    if args.BLOCK:
        print "Locked out"
        import sys; sys.exit()

    store = h5py.File(STORE_FN, "w")

    for run, orders in RUNS.iteritems():
        print run

        run_group = store.create_group(run)

        fn_results = run+"_results.dict"
        fn_exp = run+"_exp.dict"

        results = pickle.load(open(fn_results, 'r'))
        exp = pickle.load(open(fn_exp, 'r'))

        for level, folder in results.iteritems():
            print "   ", level
            level_val = int(level.split("_")[-1])

            path_to_output = os.path.join("save", folder)
            pce = pickle.load(open(os.path.join(path_to_output, "pce.p"), 'r'))['Smax']
            ## TODO: `pce` picks from a dictionary which has keys for the chaos expansions
            ##       derived each for Smax, Neq, Nkn; should add additional levels and output
            ##       to save the Neq and Nkn expansions

            level_group = run_group.create_group(level)
            coeffs = level_group.create_dataset("coeffs", pce.A.shape, dtype=pce.A.dtype)
            coeffs[:] = pce.A[:]

            orders = level_group.create_dataset("orders", pce.orders.shape,
                                                dtype=pce.orders.dtype)
            orders[:] = pce.orders[:]

            max_orders = level_group.create_dataset("max_orders", pce.max_orders.shape,
                                                    dtype=pce.max_orders.dtype)
            max_orders[:] = np.max(pce.orders, axis=0)

            stacked = np.hstack([coeffs[:][:, np.newaxis], orders[:]])
            np.savetxt("%s_%d.ascii" % (run, level_val), stacked)

    store.close()
